clear all;
close all;
%% Here, we only generate a 2-dim toy data for testing.
visualize = true;
% last two parameters are only used for visualization.
[train_data, test_data, classes, colors_samples, colors_classifiers] = get_toy_data(visualize);

%% Normalize data and initialize parameters to be trained
[dim, N] = size(train_data);

X = train_data(1:dim, :)';
y = train_data(end, :)';
%Normalize data
X = X/max(X(:));

W = rand(dim, classes);
b = zeros(1, classes);
%% Define optimization parameters
max_iter= 2000;
lambda1 = 0.01;
lambda2 = 0.02;

%compute the stepsize
L = norm(X,'fro')^2/N + lambda1;
tau = 1/L;

%% Define the functions for classifier and loss
% compute classifier scores
classifier = @(W, b, X) X*W + repmat(b, size(X, 1), 1);

% evaluate loss given classifier scores
scores = @(classifier) exp(classifier) ./ repmat(sum(exp(classifier), 2), 1, classes);
loss = @(scores) sum(-log(scores(sub2ind(size(scores), (1:N)', y))), 1);

objective = @(W, b) loss(scores(classifier(W, b, X)))/N ...
    + 0.5 * lambda1 * sum(sqrt(sum(W.^2,2)))...
    + 0.5 * lambda2 * sum(b(:).^2);

grad_W = @(W, scores, j) ...
    sum(X.*repmat(scores(:, j) - double(y == j), 1, dim), 1)'/N;

grad_b = @(b, scores, j) ...
    sum(scores(:, j) - double(y == j), 1)'/N  ...
    + lambda2 * b(j);
%% Allocate memory for storing energy etc.
obj_val = zeros(1, max_iter);
test_error = zeros(1, max_iter);
train_error = zeros(1, max_iter);

%% Proximal gradient here.
for iter=1:max_iter
    scores_train = scores(classifier(W, b, X));
    scores_test = scores(classifier(W, b, test_data(1:dim,:)'));
    
    [~, predicted_classes_train] = max(scores_train');
    [~, predicted_classes_test] = max(scores_test');
    
    % compute current objective value, test error and training error
    obj_val(iter) = objective(W, b);
    train_error(iter) = nnz(predicted_classes_train'~=train_data(end,:)')/size(train_data,2)*100;
    test_error(iter) = nnz(predicted_classes_test'~=test_data(end,:)')/size(test_data,2)*100;
  
    % print stuff
    fprintf('It=%d   obj=%f   train_error=%2.2f%%   test_error=%2.2f%%\n', ...
        iter, obj_val(iter), train_error(iter), test_error(iter));
    
    % do a gradient descent step on the differentiable part of the
    % objective.
    for j=1:classes   
        W(:, j) = W(:, j) - tau*grad_W(W, scores_train, j);
        b(j) = b(j) - tau*grad_b(b, scores_train, j);
    end
    
    % Apply proximal operator of 1,2-norm to W:
    row_norm = sqrt(sum(W.^2,2));
    W(row_norm<=tau*lambda1/2,:) = 0;
    W(row_norm>tau*lambda1/2,:) = W(row_norm>tau*lambda1/2,:)...
                                  ./row_norm(row_norm>tau*lambda1/2)...
                                  .*(row_norm(row_norm>tau*lambda1/2)-tau*lambda1/2);

    
    % evalualte stopping criterion
    if(iter > 1 && obj_val(iter-1) - obj_val(iter) < 1e-12)
        break;
    end
end

%% Plot objective value and test error
figure;
subplot(1,3,1);
plot(obj_val(1:iter));
title('Objective value');
subplot(1,3,2);
plot(train_error(1:iter));
title('Train error');
subplot(1,3,3);
plot(test_error(1:iter));
title('Test error');


%% Plot classifiers on test data
s = 0.1;
[x1Grid, x2Grid] = meshgrid(-4:s:4,...
    -4:s:4);
xGrid = [x1Grid(:),x2Grid(:)];
scores_train = classifier(W(1:2,:), b, xGrid);
figure;
hold on
for j=1:classes
    scores_test = scores(classifier(W, b, test_data(1:dim,:)'));
    [~, predicted_classes_test] = max(scores_test');
    
    scatter(test_data(1, predicted_classes_test==j),...
        test_data(2, predicted_classes_test==j),colors_samples{j});
    
    [~,h] = contour(x1Grid,x2Grid,reshape(scores_train(:, j),size(x1Grid)),[0 0],'k');
    h.LineColor = colors_classifiers{j};   
end
drawnow
title('Predicted classes for test data and classifiers')